// Copyright (C) [2009] [Robin und Detlef Wolf]
// This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 3 of the License, or (at your option) any later version.
// This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// You should have received a copy of the GNU General Public License along with this program; if not, see <http://www.gnu.org/licenses/>.

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.image.ImageObserver;
import java.util.Iterator;

import javax.swing.JPanel;

@SuppressWarnings("serial")
public class VRBahn extends JPanel implements VRObserver,
																							MouseMotionListener, 
                                              MouseListener, 
                                              ComponentListener {
  VRModel m ;
    
  public VRBahn(VRModel m) {
  	this.m = m ;
  	m.addObserver((VRObserver)this) ;
  	setPreferredSize(new Dimension(1015,620)) ;
  	addMouseMotionListener(this);
  	addMouseListener(this) ;
  	addComponentListener(this);
  	handleResize() ;
  
  }

  int pointi = 0 ;  // block indicated by mouse cursor
  int pointj = 0 ;
  boolean pointIsSelected = false ;
  int pointsi = 0 ;  // block selected by mouse cursor
  int pointsj = 0 ;
  
  int height ;  // of this panel in pixels
  int width ;   // of this panel in pixels
  private float bw ;  // width of a block in pixels
  private float bh ;  // height of a block in pixels
  
  private void handleResize() {
  	height = getHeight() ;
    width = getWidth() ;
    bw = (float)(width-10) / (float)VRModel.numDotsX ; // block width
	  bh = (float)(height-10) / (float)VRModel.numDotsY ; // block height
  }
  
  // ----------------- component listener
  
	public void componentHidden(ComponentEvent arg0) {}
	
	public void componentMoved(ComponentEvent arg0) {}

	public void componentResized(ComponentEvent arg0) {
		handleResize() ;
	}
	
	public void componentShown(ComponentEvent arg0) {}
	
  // -------------------- mouse motion listener 
	
  public void mouseMoved(MouseEvent e) {
   	pointi = x2i(e.getX()) ;
	  pointj = y2j(e.getY()) ;
	  pointIsSelected = false ;
    repaint() ;
  }
  
  public void mouseDragged(MouseEvent e) {
  	int i = x2i(e.getX()) ;
	  int j = y2j(e.getY()) ;
	  if (m.isRaceMode) {
  	  if (i != pointsi || j != pointsj)
	    	pointIsSelected = false ;
  	  else
  	  	pointIsSelected = true ;
	  }
	  else {
	  	m.addBlock(new Point(i,j)) ;
	  	pointIsSelected = false ;
	  }
    repaint() ;
  }
  
  // ------------------- mouse listener 
  
  public void mouseEntered(MouseEvent e) {
  	;
  }
  public void mouseExited(MouseEvent e) {
  	;
  }
  public void mousePressed(MouseEvent e) {
  	pointi = x2i(e.getX()) ;
	  pointj = y2j(e.getY()) ;
	  pointsi = pointi ;
	  pointsj = pointj ;
	  if (m.isRaceMode) 
	  	pointIsSelected = true ;
    repaint() ;
  }
  public void mouseReleased(MouseEvent e) {
  	pointi = x2i(e.getX()) ;
  	pointj = y2j(e.getY()) ;
  	if (!m.isRaceMode) {
  		m.addBlock(new Point(pointi,pointj)) ;
  	}
  	else if (pointIsSelected) {
    	Vektor v = m.getLastVektor() ;
  		m.addVektor(new Vektor(v.x+v.dx, v.y+v.dy,
  			pointi - (v.x+v.dx), pointj - (v.y+v.dy))) ;
  		pointIsSelected = false ;
   	}
    repaint() ;
  }
  public void mouseClicked(MouseEvent e) {
  	;
  }
  
  // -------------------- end of listeners
  
  private int x2i(int x) { // convert pixel into block#
	  return (int)((float)x / (float)bw + 0.5F) ;
  }
  
  private int y2j(int y) { // convert pixel into block#
	  return (int)((float)(height-y) / (float)bh + 0.5F) ;
  }
  
  private int i2x(float i) { // convert block coordinate into pixel
	  return (int)(i * bw) + 5; 
  }
  
  private int j2y(float j) { // convert block coordinate into pixel
	  return height - (int)(j * bh) - 5 ; 
  }
	
  // ----------
  
  private int i2x(int i) { // convert block# into pixel
  	return i2x((float)i) ;
  }
  
  private int j2y(int j) { // convert block# into pixel
  	return j2y((float)j) ;
  }
	  
    
  public void paintComponent(Graphics g) {
  	super.paintComponent(g);
  	int i ;
  	int j ;
  	
  	// draw finish line
  	g.setColor(Color.green) ;
  	g.drawLine(width / 2, (int)(height- 1.5 * bh), width / 2, (int)(height - 4.5 * bh)) ;
  	
  	// draw race track
  	Point block ;
  	Iterator<Point> bi = m.blocks.iterator() ;
  	g.setColor(Color.GRAY) ;
  	while (bi.hasNext()) {
  		block = bi.next();
  		g.fillRect(i2x(block.x)-1, j2y(block.y+1)-1, (int)(bw+0.5F)+2, (int)(bh+0.5F)+2);
  	}
  	
  	// draw grid dots
  	g.setColor(Color.MAGENTA);
  	for (i=0; i<VRModel.numDotsX; ++i) {
  		for (j=0; j<VRModel.numDotsY; ++j) {
  			g.fillOval(i2x(i)-2, j2y(j)-2, 4, 4) ; 
  		}
  	}
  	
  	// draw vectors of each player in alternating order
  	Player p ;
  	Vektor v ;
  	Boolean atEnd = false ;
  	i = 0 ;
  	while (!atEnd) {
  		Iterator<Player> pi = m.players.iterator() ;
  		atEnd = true ;
    	while (pi.hasNext()) {
    		p = pi.next();
    		if (i<p.path.size()) {
    			v = p.path.get(i) ;
    			g.setColor(p.getColor()) ;
    			drawArrow(g, i2x(v.x), j2y(v.y), 
		   	      i2x(v.x + v.dx), 
		   	      j2y(v.y + v.dy));
    			atEnd = false ;
    		}
    	}
    	i = i + 1 ;
  	}
  	
  	// draw cars at end points
  	Iterator<Player> pi = m.players.iterator() ;
		while (pi.hasNext()) {
  		p = pi.next();
  		v = p.path.get(p.path.size()-1) ;
  		p.car.draw((Graphics2D) g, i2x(v.x+v.dx), j2y(v.y+v.dy) 
  				, v.getAngle((float)width / (float)VRModel.numDotsX, 
  					  	 			 (float)height / (float)VRModel.numDotsY) 
  				, (ImageObserver)this) ;
  	}
  	
		// from debugging time
	  // Car car = m.players.get(0).car ;
	  // for (int k=0; k<=16; ++k) 
	  //   car.draw((Graphics2D) g, i2x(3), j2y(k+3), Math.toRadians(2*k), (ImageObserver)this) ;
			
		
		// draw crash points
  	Iterator<Player> pi2 = m.players.iterator() ;
		while (pi2.hasNext()) {
  		p = pi2.next();
  		Iterator<PointFloat> cpi = p.crashPoints.iterator() ;
  		while (cpi.hasNext()) {
  			PointFloat cp = cpi.next() ;
        // in case image is not available, draw substitute
    		for (i=2; i<10; i=i+2) {
     		  g.setColor(Color.red) ;
    		  g.drawOval(i2x(cp.x)-i, j2y(cp.y)-i, i*2, i*2) ;
    		}
    		m.carCrash.draw((Graphics2D) g, i2x(cp.x), j2y(cp.y), (ImageObserver)this) ;
  		}
		}  	
  	
  	if (m.isRaceMode) {  // show possible points of current player
  		Point pp ;	
  		Iterator<Point> ppi = m.getCurrentPlayer().possiblePoints.iterator() ;
  		g.setColor(m.getCurrentPlayer().getColor()) ;
  		while (ppi.hasNext()) {
  			pp = ppi.next();
  			g.drawOval(i2x(pp.x)-6, j2y(pp.y)-6, 12, 12) ;
  		}
  	}
  	
  	if (pointIsSelected) {
  		g.setColor(Color.ORANGE) ;
  		g.drawOval(i2x(pointi)-4, j2y(pointj)-4, 8, 8) ;
  		g.drawOval(i2x(pointi)-5, j2y(pointj)-5, 10, 10) ;
  		g.drawOval(i2x(pointi)-6, j2y(pointj)-6, 12, 12) ;
  	}
  	g.setColor(Color.BLACK) ;
  	g.drawOval(i2x(pointi)-3, j2y(pointj)-3, 6, 6) ;
    
  }
  
  private void drawArrow(Graphics g, int x0, int y0, int x1,int y1){
  	int deltaX = x1 - x0;
  	int deltaY = y1 - y0;
  	double frac = 0.1;
  	// copied from
  	// http://lifshitz.ucdavis.edu/~dmartin/teach_java/slope/arrows.html
	
  	g.drawLine(x0,y0,x1,y1);
  	g.drawLine(x0 + (int)((1-frac)*deltaX + frac*deltaY),
			y0 + (int)((1-frac)*deltaY - frac*deltaX),
			x1, y1);
  	g.drawLine(x0 + (int)((1-frac)*deltaX - frac*deltaY),
			y0 + (int)((1-frac)*deltaY + frac*deltaX),
			x1, y1);
  }

	public void modelChanged() {
		repaint() ;		
	}

}
