// Copyright (C) [2009] [Robin und Detlef Wolf]
// This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 3 of the License, or (at your option) any later version.
// This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// You should have received a copy of the GNU General Public License along with this program; if not, see <http://www.gnu.org/licenses/>.

import java.awt.FileDialog;
import java.awt.Point;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Iterator;

import javax.swing.JFrame;

public class VRFile {
	
	JFrame parent = null ;

	private class TextFileFilter implements FilenameFilter {
		public boolean accept(File dir, String name) {
			return name.endsWith(".txt") || name.endsWith(".TXT") ;
		}
	}
	
	public VRFile(JFrame f) { // if open/save dialogs are used
		parent = f ;
	}

	public VRFile() { // if no open/save dialogs are used
		;
	}
	
	
	private String selectFile(String title, int mode) {
		// mode: FileDialog.LOAD or FileDialog.SAVE
		String directory = System.getProperty("user.home" + File.pathSeparator + "Desktop");
		FileDialog d = new FileDialog(parent, title, mode);
		TextFileFilter fnf = new TextFileFilter() ;
		d.setDirectory(directory);       // Set the default directory
		d.setFilenameFilter(fnf) ;  // From Java Doc: Filename filters do not function in Sun's reference implementation for Microsoft Windows. 
		d.setVisible(true);
		String fileName = d.getFile();
		if (fileName == null) 
			return null ;
		directory = d.getDirectory();    // Remember new default directory
		System.out.println("You selected: " + fileName + " directory "+directory);
		return directory + File.separatorChar + fileName ;
	}
	
	private void loadFile(VRModel m, URL u) {
		BufferedReader input = null ;
    try {
    	input = new BufferedReader(new InputStreamReader(
  					u.openStream())) ;
      loadFromBuffer(m, input) ;
    }
    catch (IOException ex) {
      ex.printStackTrace();
    }

    if (input != null)
    	try {
    		input.close();
    	}
    	catch (IOException ex) {
        System.err.println("could not close" + u);
      }
	}

	public void loadFile(VRModel m, String fileName) {
		// load from 'fileName', prefer .jar file
		URL u ;
		u = VRModel.class.getResource(fileName); // read from .jar file
		if (u == null) {
			try {
				u = new URL("file:" + fileName) ;
			}
			catch ( MalformedURLException e) {
		    System.out.println("cannot find " + fileName) ;
			}
		}
		loadFile(m, u) ;
	}

	public void loadFile(VRModel m) {
		// let user select file, then load from it. 
		// Do not consider .jar file
		String fileName = selectFile("Lade Rennbahn", FileDialog.LOAD) ;
		if (fileName == null)
			return ;
		URL u = null ;
		try {
			u = new URL("file:" + fileName) ;
		}
		catch ( MalformedURLException e) {
	    System.out.println("cannot construct URL for " + fileName) ;
		}
		loadFile(m, u) ;
	}

	public void loadFromBuffer(VRModel m, BufferedReader input) throws IOException {
		int lineNo = 0 ;
		String line = null; 
		m.removeAllBlocks() ;
		while (( line = input.readLine()) != null) {
			// System.out.println(line);
			parseLine(++lineNo, line, m) ;
    }	
		m.modelChanged() ;
  }
	
	private void parseLine(int lineNo, String line, VRModel m) {
		if (line.startsWith("#"))
			return ;
		Point p = new Point() ;
		String[] f = line.split("\\s+") ;
		if (f.length >=1) {
			if (f[0].equals("block")) {
				if (f.length == 3) {
					p.x = Integer.parseInt(f[1]) ;
					p.y = Integer.parseInt(f[2]) ;
					m.addBlock(p) ;
				}
				else {
					System.out.println("line " + lineNo + ": unrecognized block line format: " + line);
				}
			}
			else {
				System.out.println("line " + lineNo + ": unrecognized line type: " + line);
			}
		}
		
	}

	public void saveFile(VRModel m) {
		String fileName = selectFile("Speichere Rennbahn", FileDialog.SAVE) ;
		if (fileName == null)
			return ;
		if (!fileName.matches("(?i)\\.TXT$")) {
			System.out.println("filename " + fileName + " does not end in .txt") ;
			fileName = fileName.concat(".txt") ;
		}
		Point block ;
  	
		try {
      BufferedWriter output = new BufferedWriter(new FileWriter(new File(fileName)));
      try {
      	Iterator<Point> bi = m.blocks.iterator() ;
      	while (bi.hasNext()) {
      		block = bi.next();
      		output.write("block " + block.x + " " + block.y) ;
      		output.newLine() ;
      	}
      }
      finally {
        output.close();
      }
    }
    catch (IOException ex){
      ex.printStackTrace();
    }
	}

}
