// Copyright (C) [2009] [Robin und Detlef Wolf]
// This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 3 of the License, or (at your option) any later version.
// This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// You should have received a copy of the GNU General Public License along with this program; if not, see <http://www.gnu.org/licenses/>.

import java.awt.Color;
import java.awt.Image;
import java.awt.Point;
import java.awt.Toolkit;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.ListIterator;

public class VRModel {
	public boolean isApplet = false ;
  public static final int numDotsX = 36 ;
  public static final int numDotsY = 22 ;
  ArrayList<Player> players = new ArrayList<Player>() ;
  private int numPlayers = 2 ; // current number of players
	int who ; // 0: player 1, 1 = player 2, [2 = player 3]
  boolean isRaceMode = false ;
  ArrayList<Point> blocks = new ArrayList<Point>() ;
  CarCrash carCrash = null ;
  ArrayList<VRObserver> observers = new ArrayList<VRObserver>() ;
  public static final String DEFAULT_TRACK = "tracks/vrstandard_hallo.txt" ;
  
  private Image fetchImage(String name) {
  	URL u ;
  	Image i = null ;
  	u = VRModel.class.getResource(name); // read from .jar file
  	if (u != null) 
  		i = Toolkit.getDefaultToolkit().getImage(u) ;
  	else
  		i = Toolkit.getDefaultToolkit().getImage(name) ;
  	return i ;
  }
  
  public VRModel() {
  	Image crashImage = fetchImage("images/crash_mini.gif") ;
  	carCrash = new CarCrash(crashImage) ;
  	players.add(new Player(
  			          new Car(Color.blue, 
  			          		    fetchImage("images/bluecar.gif")))) ;
  	players.add(new Player(
  			          new Car(Color.red, 
  			          		    fetchImage("images/redcar.gif")))) ;
  	who = 0 ;
  	setInitialVektors();
   	
  	setBorder() ; // set outer border
  }

	public void setNumPlayers(int n) {
		if (numPlayers == n) {
			System.out.println("setNumPlayers: no change.") ;
			return ;
		}
		if (numPlayers == 2 && n == 3) {
			players.add(new Player(
					new Car(new Color(167, 162, 51), 
							fetchImage("images/yellowcar.gif")))) ;
			who = 2 ;
			addVektor(new Vektor(numDotsX / 2,4,1,0)) ;
		}
		else if (numPlayers == 3 && n == 2) {
			players.remove(2) ;
			who = 0 ;
		}
		else
			System.err.println("VRModel.setNumPlayers -- unexpected usage") ;
		numPlayers = n ;
		playAgain() ;
	}

	private void setInitialVektors() {
		addVektor(new Vektor(numDotsX / 2,2,1,0)) ;
  	addVektor(new Vektor(numDotsX / 2,3,1,0)) ;
  	if (numPlayers == 3)
  		addVektor(new Vektor(numDotsX / 2,4,1,0)) ;
  }
  
  
  private void setBorder() {
  	for (int i=0; i<numDotsX; ++i)
  		addBlock(new Point(i, numDotsY-1)) ;
  	for (int i=0; i<numDotsX; ++i)
  		addBlock(new Point(i, 0)) ;
  	for (int j=0; j<numDotsY; ++j)
  		addBlock(new Point(0, j)) ;
  	for (int j=0; j<numDotsY; ++j)
  		addBlock(new Point(numDotsX-1, j)) ;
  }
  
  public void playAgain() {
  	Iterator<Player> pi = players.iterator() ;
		while (pi.hasNext()) 
  		pi.next().clear();
  	who = 0 ;
  	setInitialVektors();
  	modelChanged() ;
 	}
  
  public void clearTrack() {
  	isRaceMode = false ;
    blocks.clear();
  	setBorder() ;
  	playAgain() ; // triggers modelChanged() ;
  }
  
  public void modelChanged() {
  	Iterator<VRObserver> oi = observers.iterator() ;
    while (oi.hasNext()) 
  		oi.next().modelChanged();
  }
 	
  public Player getCurrentPlayer() {
  	return players.get(who) ;
  }
  
  public Vektor getLastVektor() {  // get last Vektor of current player
  	Player p = players.get(who) ;
  	return p.path.get(p.path.size()-1) ;
  }
  
  public void addVektor(Vektor v) {
  	// 1) current player makes move v
  	// 2) advance to next player
  	Player p = players.get(who) ;
  	PointFloat crashPoint = findCrash(v) ; 
   	p.add(v) ;
   	if (crashPoint != null) {
 		  sound.play() ;
 		  p.addCrashPoint(crashPoint) ;
    }
   	{ // detect crash with other player
   		for (int k = 0; k<numPlayers; ++k) {
   			if (k == who)
   				continue ;
   			Player other = players.get(k) ;
   			if (other.path.size() > 0) {
   				Vektor v2 = other.path.get(other.path.size()-1) ;
   				if (v2.hasSameEndPoint(v)) {
   					PointFloat crashPoint2 = new PointFloat(v2.x+v2.dx, v2.y+v2.dy) ;
   					sound.play() ;
   					p.addCrashPoint(crashPoint2) ;
   				}
   			}
   		}
   	}
   	
   	who = (who+1) % numPlayers ;
  }

  private PointFloat findCrash(Vektor v) {
  	// returns point where v crashes, otherwise null
  	Point block ;
  	PointDistance blockDistance ;
  	int distance2 ; // square of distance
  	int dx;
  	int dy;
    PointFloat crashPoint = null ;
  	ArrayList<PointDistance> blocksDistance = new ArrayList<PointDistance>() ;
  	Iterator<Point> bi = blocks.iterator() ;
  	while (bi.hasNext()) {
  		block = bi.next() ;
  		dx= v.x-block.x ;
  		dy= v.y-block.y ;
  		distance2 = dx*dx+dy*dy ; 
  	  PointDistance pointDistance = new PointDistance(block, distance2) ;
  	  blocksDistance.add(pointDistance);
  	}
  	Collections.sort(blocksDistance) ;
  	
  	Iterator<PointDistance> bdi = blocksDistance.iterator() ;
  	while (bdi.hasNext()) {
  		blockDistance = bdi.next() ;
  	  // System.out.println(blockDistance);
  		crashPoint = intersects(v, blockDistance.point) ;
  		if (crashPoint != null)
  			return crashPoint ;
  	}
  	return null ;
  }

	private PointFloat intersects(Vektor v, Point block) {
		// returns point where v intersects block if such a point exists,
		// null otherwise
		PointFloat crashPoint = null ;
		if (v.dx != 0) {  // only if vector is not vertical
			// check for intersection with left wall
			if (v.x <= block.x && v.dx >= block.x - v.x) {
				// only if vector starts left of block and
				// is long enough to reach the block
				float ys = (float)v.y + (float)v.dy / (float)v.dx 
				         * ((float)block.x - (float)v.x) ;
				if (ys >= (float)block.y && ys <= (float)block.y + 1.0F) {
					System.out.println("crash left " + block + v) ;
					crashPoint = new PointFloat((float)block.x, ys) ;
					return crashPoint ;
				}
			}
			
			if (v.x >= block.x+1 && v.x - (block.x +1) + v.dx <= 0) {
				// only if vector starts right of block and
				// is long enough to reach the block
				float ys = (float)v.y + (float)v.dy / (float)v.dx 
				         * ((float)(block.x+1) - (float)v.x) ;
				if (ys >= (float)block.y && ys <= (float)block.y + 1.0F) {
					System.out.println("crash right " + block + v) ;
					crashPoint = new PointFloat((float)(block.x+1), ys) ;
					return crashPoint ;
				}
			}
		}
		
		if (v.dy != 0) {
			if (v.y >= block.y+1 && v.y - (block.y +1) + v.dy <= 0) {
				// only if vector starts above of block and
				// is long enough to reach the block
				float xs = (float)v.x + (float)v.dx / (float)v.dy * (float)(block.y+1 - v.y) ;
				if (xs >= (float)block.x && xs <= (float)(block.x +1)) {
					System.out.println("crash top " + block + v + " @" +xs) ;
					crashPoint = new PointFloat(xs, (float)(block.y +1)) ;
					return crashPoint ;
				}
			}
			
			if (v.y <= block.y && v.dy >= block.y - v.y) {
				// only if vector starts below block and
				// is long enough to reach the block
				float xs = (float)v.x + (float)v.dx / (float)v.dy * (float)(block.y - v.y) ;
				if (xs >= (float)block.x && xs <= (float)(block.x +1)) {
					System.out.println("crash bottom " + block + v + " @" +xs) ;
					crashPoint = new PointFloat(xs, (float)(block.y)) ;
					return crashPoint ;
				}
			}
			
		}
		return null ;
	}

	public void addBlock(Point p) {
  	boolean found = false ;
  	ListIterator<Point> bi = blocks.listIterator(blocks.size()) ;
  	while (bi.hasPrevious()) {
  		Point block = bi.previous();
  		if (block.equals(p)) {
  			found = true ;
  			break ;
  		}
  	}
  	if (!found) {
  		blocks.add(p) ;
  	}
  }
  
  public void removeAllBlocks() {
  	blocks.clear() ;
	}

  public void setRaceMode(boolean isRaceMode) {
		this.isRaceMode = isRaceMode ;
		modelChanged() ;
	}
  
  public void addObserver(VRObserver observer) {
  	observers.add(observer) ;
  }

  private WavPlayer sound = null ;
	public void setSound(WavPlayer sound) {
		this.sound = sound ;
	}

	public WavPlayer getSound() {
		return sound ;
	}

	
}
